@@ -32,12 +32,14 @@ const char *version = VERSION_STRING;	/* Program version */
 static pid_t my_pgrp;		        /* The "magic" process group */
 static pid_t my_pid;		        /* The pid of this process */
 
+int kproto_version;		/* Kernel protocol version */
+
 static int mounted = 0;
 
 struct pending_mount {
 	pid_t pid;		/* Which process is mounting for us */
 	unsigned long wait_queue_token;	/* Associated kernel wait token */
-	struct pending_mount *next;
+	volatile struct pending_mount *next;
 };
 
 static struct autofs_point {
@@ -45,20 +47,64 @@ static struct autofs_point {
 	int pipefd;		/* File descriptor for pipe */
 	int ioctlfd;		/* File descriptor for ioctls */
 	dev_t dev;		/* "Device" number assigned by kernel */
-	struct pending_mount *mounts; /* Pending mount queue */
+	unsigned exp_timeout;	/* Timeout for expiring mounts */
+	unsigned exp_runfreq;	/* Frequency for polling for timeouts */
+	volatile pid_t exp_process; /* Process that is currently expiring */
+	volatile struct pending_mount *mounts; /* Pending mount queue */
 	struct lookup_mod *lookup; /* Lookup module */
 } ap;
 
-static int umount_all(void)
+#define DEFAULT_TIMEOUT (5*60)	/* 5 minutes */
+#define CHECK_RATIO     4	/* exp_runfreq = exp_timeout/CHECK_RATIO */
+
+/* Used by subprocesses to avoid doing silly things with signals */
+static void reset_signals(void)
 {
-	int left, rv;
+	struct sigaction sa;
+	int i;
+
+	sa.sa_handler = SIG_DFL;
+	sigemptyset(&sa.sa_mask);
+	sa.sa_flags = 0;
+	for ( i = 1 ; i < NSIG ; i++ )
+		sigaction(i, &sa, NULL);
+}
+
+static int umount_ent(char *root, char *name)
+{
+	char path_buf[PATH_MAX];
 	struct stat st;
+	int rv = 0;
+
+	sprintf(path_buf, "%s/%s", root, name);
+	if ( !lstat(path_buf,&st) ) {
+		if ( S_ISDIR(st.st_mode) ) {
+			if ( st.st_dev != ap.dev ) {
+				rv = spawnl(_PATH_UMOUNT, _PATH_UMOUNT,
+					    path_buf, NULL);
+				if ( !rv )
+					rmdir(path_buf);
+			} else
+				rmdir(path_buf);
+		} else 
+			unlink(path_buf);
+	}
+	return rv;
+}	
+
+static int umount_all(int force)
+{
+	int left;
 	struct dirent *de;
 	DIR *dp;
-	char path_buf[PATH_MAX];
 
 	chdir("/");
 
+	if ( ap.exp_process ) {
+		kill(ap.exp_process, SIGTERM);
+		usleep(100000);
+	}
+
 	dp = opendir(ap.path);
 	if ( !dp ) {
 		syslog(LOG_ERR, "umount_all: opendir: %m");
@@ -69,21 +115,12 @@ static int umount_all(void)
 
 	while ( (de = readdir(dp)) != NULL ) {
 		if ( strcmp(de->d_name,".") && strcmp(de->d_name,"..") ) {
-			sprintf(path_buf, "%s/%s", ap.path, de->d_name);
-			if ( !lstat(path_buf,&st) ) {
-				if ( S_ISDIR(st.st_mode) ) {
-					if ( st.st_dev != ap.dev ) {
-						rv = spawnl(_PATH_UMOUNT, _PATH_UMOUNT, path_buf, NULL);
-						if ( rv ) {
-							syslog(LOG_NOTICE, "could not unmount %s", path_buf);
-							left++;
-						}
-						else
-							rmdir(path_buf);
-					} else
-						rmdir(path_buf);
-				} else 
-					unlink(path_buf);
+			if ( umount_ent(ap.path,de->d_name) ) {
+				if ( force )
+					syslog(LOG_WARNING,
+					       "could not unmount %s/%s",
+					       ap.path,de->d_name);
+				left++;
 			}
 		}
 	}
@@ -99,7 +136,7 @@ static int umount_autofs(int force)
 	if ( !mounted )
 		return -1;
 
-	if ( umount_all() && !force )
+	if ( umount_all(force) && !force )
 		return -1;
 
 	if (ap.ioctlfd >= 0) {
@@ -146,8 +183,8 @@ static int mount_autofs(char *path)
 	if ( pipe(pipefd) < 0 )
 		return -1;
 	
-	sprintf(options, "fd=%d,pgrp=%u,minproto=%d,maxproto=%d", pipefd[1],
-		(unsigned)my_pgrp, AUTOFS_PROTO_VERSION, AUTOFS_PROTO_VERSION);
+	sprintf(options, "fd=%d,pgrp=%u,minproto=2,maxproto=%d", pipefd[1],
+		(unsigned)my_pgrp, AUTOFS_PROTO_VERSION);
 	sprintf(our_name, "automount(pid%u)", (unsigned)my_pid);
 
 	if ( spawnl(_PATH_MOUNT, _PATH_MOUNT, "-t", "autofs", "-o",
@@ -177,8 +214,42 @@ static int mount_autofs(char *path)
 	return 0;
 }
 
+static void run_expire(void)
+{
+	pid_t f;
+	struct autofs_packet_expire pkt;
+
+	if ( ap.exp_process )	/* Another process is already expiring */
+		return;		/* Try again next interval */
+	
+	f = fork();
+	if ( f < 0 )
+		return;
+	if ( f > 0 ) {
+		ap.exp_process = f;
+		return;
+	}
+
+	/* This is the actual expire run, run as a subprocess */
+
+	reset_signals();
+	close(ap.pipefd);
+
+	syslog(LOG_DEBUG, "running expiration on path %s", ap.path);
+
+	while ( ioctl(ap.ioctlfd, AUTOFS_IOC_EXPIRE, &pkt) == 0 ) {
+		if ( pkt.hdr.type == autofs_ptype_expire ) {
+			if ( !umount_ent(ap.path,pkt.name) )
+				syslog(LOG_NOTICE, "expired %s/%s", ap.path,
+				       pkt.name);
+		}
+	}
+	_exit(0);
+}
+
 static sig_atomic_t shutdown = 0;
 static sig_atomic_t prune = 0;
+static sig_atomic_t expire_now = 0;
 
 static void sig_shutdown(int sig)
 {
@@ -190,6 +261,11 @@ static void sig_prune(int sig)
 	prune = 1;
 }
 
+static void sig_expire(int sig)
+{
+	expire_now = 1;
+}
+
 static int get_pkt(int fd, struct autofs_packet_missing *pkt)
 {
 	char *buf = (char *) pkt;
@@ -202,9 +278,14 @@ static int get_pkt(int fd, struct autofs_packet_missing *pkt)
 		if (i < 0) {
 			if (errno == EINTR) {
 				if ( prune ) {
-					umount_all();
+					umount_all(0);
 					prune = 0;
-				} 
+				}
+				if ( expire_now ) {
+					run_expire();
+					expire_now = 0;
+					alarm(ap.exp_runfreq);
+				}
 				if ( !shutdown )
 					continue;
 			}
@@ -263,14 +344,10 @@ static int handle_packet(void)
 			return 1;
 		} else if ( !f ) {
 			int i;
-			struct sigaction sa;
+
+			reset_signals();
 			close(ap.pipefd);
 			close(ap.ioctlfd);
-			sa.sa_handler = SIG_DFL;
-			sigemptyset(&sa.sa_mask);
-			sa.sa_flags = 0;
-			for ( i = 1 ; i < NSIG ; i++ )
-				sigaction(i, &sa, NULL);
 
 			/* Wait for go-ahead from main process */
 			kill(getpid(), SIGSTOP);
@@ -310,21 +387,27 @@ void sig_child(int sig)
 {
 	pid_t pid;
 	int status;
-	struct pending_mount *mt, **mtp;
+	volatile struct pending_mount *mt, * volatile *mtp;
 
 	pid = waitpid(-1, &status, WNOHANG);
-	if ( pid > 0 ) {
-		for ( mtp = &ap.mounts ; (mt = *mtp) ; mtp = &mt->next ) {
-			if ( mt->pid == pid ) {
-				if ( !WIFEXITED(status) && !WIFSIGNALED(status) )
-					break;
-				else if ( WIFSIGNALED(status) || WEXITSTATUS(status) != 0 )
-					send_fail(mt->wait_queue_token);
-				else
-					send_ready(mt->wait_queue_token);
-				*mtp = mt->next; /* Unlink */
+	if ( !(pid > 0) )
+		return;
+	if ( pid == ap.exp_process ) {
+		ap.exp_process = 0;
+		return;
+	}
+	for ( mtp = &ap.mounts ; (mt = *mtp) ; mtp = &mt->next ) {
+		if ( mt->pid == pid ) {
+			if ( !WIFEXITED(status) && !WIFSIGNALED(status) )
 				break;
-			}
+			else if ( WIFSIGNALED(status) ||
+				  WEXITSTATUS(status) != 0 )
+				send_fail(mt->wait_queue_token);
+			else
+				send_ready(mt->wait_queue_token);
+			*mtp = mt->next; /* Remove from linked list */
+			/*** FIXME: Need to free mt from sig handler ***/
+			break;
 		}
 	}
 }
@@ -368,7 +451,7 @@ int main(int argc, char *argv[])
 
 	program = argv[0];
 
-	if ( argc < 3 ) {
+	if ( argc < 3 || geteuid() != 0 ) {
 		fprintf(stderr, "Linux automount version %s\n", version);
 		fprintf(stderr, "Usage: %s path map_type [args...]\n", program);
 		exit(1);
@@ -409,11 +492,12 @@ int main(int argc, char *argv[])
 	sigaction(SIGFPE, &sa, NULL);
 	sigaction(SIGSEGV, &sa, NULL);
 	sigaction(SIGPIPE, &sa, NULL);
-	sigaction(SIGALRM, &sa, NULL);
 	sigaction(SIGTERM, &sa, NULL);
-	sigaction(SIGSTKFLT, &sa, NULL);
 	sigaction(SIGIO, &sa, NULL);
 	sigaction(SIGPWR, &sa, NULL);
+#ifdef SIGSTKFLT
+	sigaction(SIGSTKFLT, &sa, NULL);
+#endif
 	sigaction(SIGUSR2, &sa, NULL);
 
 	sa.sa_handler = sig_child;
@@ -426,6 +510,34 @@ int main(int argc, char *argv[])
 	sa.sa_flags = 0;
 	sigaction(SIGUSR1, &sa, NULL);
 
+	sa.sa_handler = sig_expire;
+	sigemptyset(&sa.sa_mask);
+	sa.sa_flags = 0;
+	sigaction(SIGALRM, &sa, NULL);
+
+	/* If this ioctl() doesn't work, it is kernel version 2 */
+	if ( ioctl(ap.ioctlfd, AUTOFS_IOC_PROTOVER, &kproto_version) ) {
+		syslog(LOG_DEBUG, "kproto: %m");
+		kproto_version = 2;
+	}
+
+	syslog(LOG_INFO, "using kernel protocol version %d", kproto_version);
+
+	if ( kproto_version < 3 ) {
+		ap.exp_timeout = ap.exp_runfreq = 0;
+		syslog(LOG_INFO, "kernel does not support timeouts");
+	} else {
+		unsigned long timeout;
+
+		ap.exp_timeout = DEFAULT_TIMEOUT;
+		ap.exp_runfreq = ap.exp_timeout / CHECK_RATIO;
+
+		timeout = ap.exp_timeout;
+		ioctl(ap.ioctlfd, AUTOFS_IOC_SETTIMEOUT, &timeout);
+		
+		alarm(ap.exp_timeout+ap.exp_runfreq);
+	}
+
 	while ( !shutdown ) {
 		if (handle_packet() && errno != EINTR)
 			break;
