@@ -53,7 +53,8 @@ int mount_mount(char *root, char *name, int name_len, char *what,
   char *colon, **haddr, *fullpath;
   struct hostent *he;
   struct sockaddr_in saddr, laddr;
-  int sock, len, local, err;
+  int sock, local, err;
+  size_t len;
 
   colon = strchr(what, ':');
   if ( !colon ) {
